from webrockets import Connection


class TestConnectionCreation:
    def test_create_with_all_parameters(self):
        conn = Connection(
            path="/ws/chat/",
            query_string="room=general&user=alice",
            headers={
                "authorization": "Bearer token123",
                "content-type": "application/json",
            },
            cookies={"sessionid": "abc123", "csrftoken": "xyz789"},
        )

        assert conn.path == "/ws/chat/"
        assert conn.query_string == "room=general&user=alice"
        assert conn.headers == {
            "authorization": "Bearer token123",
            "content-type": "application/json",
        }
        assert conn.cookies == {"sessionid": "abc123", "csrftoken": "xyz789"}

    def test_create_with_empty_values(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={},
            cookies={},
        )

        assert conn.path == "/ws/"
        assert conn.query_string == ""
        assert conn.headers == {}
        assert conn.cookies == {}

    def test_user_is_none_by_default(self):
        conn = Connection(
            path="/ws/test/",
            query_string="",
            headers={},
            cookies={},
        )

        assert conn.user is None


class TestConnectionCookies:
    def test_get_cookie_existing(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={},
            cookies={"sessionid": "abc123", "auth_token": "jwt-token"},
        )

        assert conn.get_cookie("sessionid") == "abc123"
        assert conn.get_cookie("auth_token") == "jwt-token"

    def test_get_cookie_nonexistent(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={},
            cookies={"sessionid": "abc123"},
        )

        assert conn.get_cookie("nonexistent") is None

    def test_get_cookie_empty_cookies(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={},
            cookies={},
        )

        assert conn.get_cookie("any_cookie") is None


class TestConnectionHeaders:
    def test_get_header_existing(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={"authorization": "Bearer token", "x-custom-header": "value"},
            cookies={},
        )

        assert conn.get_header("authorization") == "Bearer token"
        assert conn.get_header("x-custom-header") == "value"

    def test_get_header_nonexistent(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={"authorization": "Bearer token"},
            cookies={},
        )

        assert conn.get_header("nonexistent") is None

    def test_get_header_empty_headers(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={},
            cookies={},
        )

        assert conn.get_header("any_header") is None

    def test_headers_are_case_sensitive(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={"authorization": "token1", "Authorization": "token2"},
            cookies={},
        )

        assert conn.get_header("authorization") == "token1"
        assert conn.get_header("Authorization") == "token2"


class TestConnectionSpecialCharacters:
    def test_path_with_special_characters(self):
        conn = Connection(
            path="/ws/chat/room-123/user_abc/",
            query_string="",
            headers={},
            cookies={},
        )

        assert conn.path == "/ws/chat/room-123/user_abc/"

    def test_query_string_with_encoded_values(self):
        conn = Connection(
            path="/ws/",
            query_string="msg=hello%20world&name=John%20Doe",
            headers={},
            cookies={},
        )

        assert conn.query_string == "msg=hello%20world&name=John%20Doe"

    def test_cookie_with_special_values(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={},
            cookies={
                "jwt": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIn0.dozjgNryP4J3jVmNHl0w5N_XgL0n3I9PlFUP0THsR8U"
            },
        )

        assert conn.get_cookie("jwt").startswith("eyJ")

    def test_header_with_unicode(self):
        conn = Connection(
            path="/ws/",
            query_string="",
            headers={"x-user-name": "用户名"},
            cookies={},
        )

        assert conn.get_header("x-user-name") == "用户名"



from webrockets import Connection, WebsocketServer

server = WebsocketServer()

echo = server.create_route("echo")


@echo.receive
def receive(conn: Connection, data: str | bytes):
    conn.send(data)


server.start()



import asyncio
import json

import pytest
from django.db import models
from webrockets import Connection, WebsocketServer
from webrockets.client import ClientConfig, InvalidStatusCode, aconnect, connect
from webrockets.django.auth import SessionAuthentication
from webrockets.test import runserver


class Message(models.Model):
    content = models.TextField()
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        app_label = "tests"
        db_table = "test_messages"


@pytest.mark.django_db(transaction=True)
class TestDjangoORM:
    def test_orm_create_in_receive_handler(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws/orm/")

        @route.receive
        def receive(conn: Connection, data: str | bytes):
            content = data if isinstance(data, str) else data.decode()
            Message.objects.create(content=content)
            count = Message.objects.count()
            conn.send(json.dumps({"created": True, "count": count}))

        with runserver(ws_server):
            with connect(f"ws://{ws_server.addr()}/ws/orm/") as ws:
                ws.send("test message")
                response = ws.recv()
                data = json.loads(response)
                assert data["created"] is True
                assert data["count"] >= 1

    def test_orm_query_in_receive_handler(self, ws_server: WebsocketServer):
        """Test querying Django models in receive handler."""
        Message.objects.create(content="existing message 1")
        Message.objects.create(content="existing message 2")

        route = ws_server.create_route("ws/orm/")

        @route.receive
        def receive(conn: Connection, data: str | bytes):
            messages = list(Message.objects.order_by("-id")[:5])
            conn.send(json.dumps([{"id": m.id, "content": m.content} for m in messages]))

        with runserver(ws_server):
            with connect(f"ws://{ws_server.addr()}/ws/orm/") as ws:
                ws.send("query")
                response = ws.recv()
                data = json.loads(response)
                assert len(data) >= 2
                assert any(m["content"] == "existing message 1" for m in data)

    def test_orm_update_in_receive_handler(self, ws_server: WebsocketServer):
        """Test updating Django models in receive handler."""
        msg = Message.objects.create(content="original")

        route = ws_server.create_route("ws/orm/")

        @route.receive
        def receive(conn: Connection, data: str | bytes):
            payload = json.loads(data)
            Message.objects.filter(id=payload["id"]).update(content=payload["content"])
            updated = Message.objects.get(id=payload["id"])
            conn.send(json.dumps({"id": updated.id, "content": updated.content}))

        with runserver(ws_server):
            with connect(f"ws://{ws_server.addr()}/ws/orm/") as ws:
                ws.send(json.dumps({"id": msg.id, "content": "updated"}))
                response = ws.recv()
                data = json.loads(response)
                assert data["content"] == "updated"

    def test_orm_delete_in_receive_handler(self, ws_server: WebsocketServer):
        """Test deleting Django models in receive handler."""
        msg = Message.objects.create(content="to be deleted")

        route = ws_server.create_route("ws/orm/")

        @route.receive
        def receive(conn: Connection, data: str | bytes):
            payload = json.loads(data)
            deleted, _ = Message.objects.filter(id=payload["id"]).delete()
            conn.send(json.dumps({"deleted": deleted}))

        with runserver(ws_server):
            with connect(f"ws://{ws_server.addr()}/ws/orm/") as ws:
                ws.send(json.dumps({"id": msg.id}))
                response = ws.recv()
                data = json.loads(response)
                assert data["deleted"] == 1

        assert not Message.objects.filter(id=msg.id).exists()

    @pytest.mark.asyncio
    async def test_async_orm_operations(self, ws_server: WebsocketServer):
        """Test async ORM operations using sync_to_async."""
        route = ws_server.create_route("ws/orm/")

        @route.receive
        async def receive(conn: Connection, data: str | bytes):
            content = data if isinstance(data, str) else data.decode()

            msg = await Message.objects.acreate(content=content)
            await conn.asend(json.dumps({"id": msg.id, "content": msg.content}))

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws/orm/") as ws:
                await ws.send("async message")
                response = await asyncio.wait_for(ws.recv(), timeout=2.0)
                data = json.loads(response)
                assert data["content"] == "async message"


@pytest.mark.django_db(transaction=True)
class TestDjangoSessionAuth:
    """Tests for Django session authentication with websocket connections."""

    def test_session_auth_valid_session(
        self, ws_server: WebsocketServer, create_session, active_user
    ):
        session = create_session(user=active_user)
        route = ws_server.create_route("ws/auth/", authentication_classes=[SessionAuthentication()])

        @route.connect("after")
        def after_connect(conn: Connection):
            conn.send(json.dumps({"user_id": conn.user.id, "username": conn.user.username}))

        with runserver(ws_server):
            with connect(
                f"ws://{ws_server.addr()}/ws/auth/",
                config=ClientConfig(
                    extra_headers={"Cookie": f"sessionid={session.session_key}"},
                ),
            ) as ws:
                response = ws.recv()
                data = json.loads(response)
                assert data["user_id"] == active_user.id
                assert data["username"] == active_user.username

    def test_session_auth_invalid_session(self, ws_server: WebsocketServer):
        """Test websocket connection rejection with invalid session."""
        route = ws_server.create_route("ws/auth/", authentication_classes=[SessionAuthentication()])

        @route.connect("after")
        def after_connect(conn: Connection):
            conn.send("connected")

        with runserver(ws_server):
            with pytest.raises(InvalidStatusCode) as exc_info:
                with connect(
                    f"ws://{ws_server.addr()}/ws/auth/",
                    config=ClientConfig(extra_headers={"Cookie": "sessionid=invalid-session"}),
                ):
                    pass

            assert exc_info.value.status_code == 401

    def test_session_auth_no_session(self, ws_server: WebsocketServer):
        """Test websocket connection rejection without session cookie."""
        route = ws_server.create_route("ws/auth/", authentication_classes=[SessionAuthentication()])

        @route.connect("after")
        def after_connect(conn: Connection):
            conn.send("connected")

        with runserver(ws_server):
            with pytest.raises(InvalidStatusCode) as exc_info:
                with connect(f"ws://{ws_server.addr()}/ws/auth/"):
                    pass

            assert exc_info.value.status_code == 401

    def test_session_auth_inactive_user(
        self, ws_server: WebsocketServer, session_store, inactive_user
    ):
        session = session_store()
        session["_auth_user_id"] = str(inactive_user.pk)
        session["_auth_user_backend"] = "django.contrib.auth.backends.ModelBackend"
        session.create()

        route = ws_server.create_route("ws/auth/", authentication_classes=[SessionAuthentication()])

        @route.connect("after")
        def after_connect(conn: Connection):
            conn.send("connected")

        with runserver(ws_server):
            with pytest.raises(InvalidStatusCode) as exc_info:
                with connect(
                    f"ws://{ws_server.addr()}/ws/auth/",
                    config=ClientConfig(
                        extra_headers={"Cookie": f"sessionid={session.session_key}"}
                    ),
                ):
                    pass

            assert exc_info.value.status_code == 401

    def test_authenticated_user_in_receive_handler(
        self, ws_server: WebsocketServer, create_session, active_user
    ):
        session = create_session(user=active_user)
        route = ws_server.create_route("ws/auth/", authentication_classes=[SessionAuthentication()])

        @route.receive
        def receive(conn: Connection, data: str | bytes):
            conn.send(
                json.dumps(
                    {
                        "received": data if isinstance(data, str) else data.decode(),
                        "from_user": conn.user.username,
                    }
                )
            )

        with runserver(ws_server):
            with connect(
                f"ws://{ws_server.addr()}/ws/auth/",
                config=ClientConfig(extra_headers={"Cookie": f"sessionid={session.session_key}"}),
            ) as ws:
                ws.send("hello")
                response = ws.recv()
                data = json.loads(response)
                assert data["received"] == "hello"
                assert data["from_user"] == active_user.username


@pytest.mark.django_db(transaction=True)
class TestDjangoCacheIntegration:
    def test_cache_set_get_in_handler(self, ws_server: WebsocketServer):
        from django.core.cache import cache

        route = ws_server.create_route("ws/cache/")

        @route.receive
        def receive(conn: Connection, data: str | bytes):
            payload = json.loads(data)
            if payload.get("action") == "set":
                cache.set(payload["key"], payload["value"], timeout=60)
                conn.send(json.dumps({"status": "set"}))
            elif payload.get("action") == "get":
                value = cache.get(payload["key"])
                conn.send(json.dumps({"value": value}))

        with runserver(ws_server):
            with connect(f"ws://{ws_server.addr()}/ws/cache/") as ws:
                ws.send(json.dumps({"action": "set", "key": "ws_test", "value": "cached"}))
                response = ws.recv()
                assert json.loads(response)["status"] == "set"

                ws.send(json.dumps({"action": "get", "key": "ws_test"}))
                response = ws.recv()
                assert json.loads(response)["value"] == "cached"



import importlib
import logging

if importlib.util.find_spec("django") is None:
    raise ImportError(
        'Django is required to use runwebsockets command. Install with: pip install "webrockets[django]"'
    )

from django.core.management.base import BaseCommand
from django.utils.module_loading import autodiscover_modules

from webrockets.django import server

LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


class Command(BaseCommand):
    help = "Start the webrockets WebSocket server"

    def add_arguments(self, parser):
        parser.add_argument(
            "--log-level",
            type=str,
            choices=LOG_LEVELS.keys(),
            default=None,
            help="Set the log level for webrockets (debug, info, warning, error, critical)",
        )

    def handle(self, *args, **options):
        if options["log_level"]:
            level = LOG_LEVELS[options["log_level"]]
            logger = logging.getLogger("webrockets")
            logger.setLevel(level)
            if not logger.handlers:
                handler = logging.StreamHandler()
                handler.setFormatter(
                    logging.Formatter("%(asctime)s %(levelname)s %(name)s: %(message)s")
                )
                logger.addHandler(handler)
                logger.propagate = False

        autodiscover_modules("websockets", "sockets", "sse", "views")
        server.start()



import asyncio
import tempfile
import threading
from collections.abc import Callable
from functools import wraps
from pathlib import Path
from typing import Any

import django
import pytest
from django.conf import settings
from django.contrib.auth import get_user_model
from webrockets import WebsocketRoute, WebsocketServer
from webrockets.test import runserver

# Use a temp file for SQLite so all threads can access it
_test_db_file = Path(tempfile.gettempdir()) / "webrockets_test.sqlite3"


class _ErrorCapture:
    def __init__(self):
        self._errors: list[BaseException] = []
        self._lock = threading.Lock()

    def wrap(self, func: Callable) -> Callable:
        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except BaseException as e:
                with self._lock:
                    self._errors.append(e)
                raise

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except BaseException as e:
                with self._lock:
                    self._errors.append(e)
                raise

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        return sync_wrapper

    def check(self):
        with self._lock:
            if self._errors:
                raise self._errors[0]


class _TestRoute:
    def __init__(self, route: WebsocketRoute, error_capture: _ErrorCapture):
        self._route = route
        self._errors = error_capture

    def connect(self, *args, **kwargs) -> Any:
        original_decorator = self._route.connect(*args, **kwargs)

        def wrapper(func):
            return original_decorator(self._errors.wrap(func))

        return wrapper

    def receive(self, *args, **kwargs) -> Any:
        if args and callable(args[0]) and not kwargs:
            # Called as @route.receive without arguments
            func = args[0]
            return self._route.receive(self._errors.wrap(func))
        else:
            # Called as @route.receive(...) with arguments
            original_decorator = self._route.receive(*args, **kwargs)

            def wrapper(func):
                return original_decorator(self._errors.wrap(func))

            return wrapper

    def disconnect(self, *args, **kwargs) -> Any:
        func = args[0]
        return self._route.disconnect(self._errors.wrap(func))

    def __getattr__(self, name):
        return getattr(self._route, name)


class TestWebsocketServer:
    def __init__(self, server: WebsocketServer | None = None) -> None:
        if server:
            self._server = server
        else:
            self._server = WebsocketServer()
        self._errors = _ErrorCapture()

    def create_route(self, *args, **kwargs) -> Any:
        route = self._server.create_route(*args, **kwargs)
        return _TestRoute(route, self._errors)

    def check_errors(self):
        self._errors.check()

    def __getattr__(self, name):
        return getattr(self._server, name)


class runtestserver(runserver):
    def __init__(self, server: WebsocketServer) -> None:
        if isinstance(server, TestWebsocketServer):
            self._server = server._server
            self._test_server = server
        else:
            self._server = server
            self._test_server = None

    def __exit__(self, exc_type, exc, tb):
        super().__exit__(exc_type, exc, tb)
        if exc_type is None and self._test_server is not None:
            self._test_server.check_errors()


def pytest_configure():
    """Configure Django settings for tests."""
    if not settings.configured:
        settings.configure(
            DEBUG=True,
            DATABASES={
                "default": {
                    "ENGINE": "django.db.backends.sqlite3",
                    "NAME": str(_test_db_file),
                    "OPTIONS": {
                        "timeout": 20,
                    },
                }
            },
            INSTALLED_APPS=[
                "django.contrib.contenttypes",
                "django.contrib.auth",
                "django.contrib.sessions",
                "webrockets",
            ],
            SECRET_KEY="test-secret-key-for-testing-only",
            SESSION_ENGINE="django.contrib.sessions.backends.db",
            SESSION_COOKIE_NAME="sessionid",
            USE_TZ=True,
            CACHES={
                "default": {
                    "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
                }
            },
        )
        django.setup()


def pytest_unconfigure():
    """Clean up test database file."""
    if _test_db_file.exists():
        _test_db_file.unlink()


@pytest.fixture(scope="session")
def django_db_setup(django_db_blocker):
    from django.core.management import call_command
    from django.db import connection

    with django_db_blocker.unblock():
        call_command("migrate", "--run-syncdb", verbosity=0)

        with connection.cursor() as cursor:
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS test_messages (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    content TEXT NOT NULL,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                )
            """)


@pytest.fixture
def user_model():
    """Return the User model."""
    return get_user_model()


@pytest.fixture
def create_user(user_model, django_db_blocker):
    """Factory fixture to create users."""

    def _create_user(username="testuser", password="testpass123", **kwargs):
        with django_db_blocker.unblock():
            return user_model.objects.create_user(username=username, password=password, **kwargs)

    return _create_user


@pytest.fixture
def active_user(create_user):
    return create_user(username="activeuser", is_active=True)


@pytest.fixture
def inactive_user(create_user):
    return create_user(username="inactiveuser", is_active=False)


@pytest.fixture
def session_store():
    from django.contrib.sessions.backends.db import SessionStore

    return SessionStore


@pytest.fixture
def create_session(session_store, active_user):
    def _create_session(user=None, expired=False):
        user = user or active_user
        session = session_store()
        session["_auth_user_id"] = str(user.pk)
        session["_auth_user_backend"] = "django.contrib.auth.backends.ModelBackend"

        # Add session auth hash if user has it
        if hasattr(user, "get_session_auth_hash"):
            session["_auth_user_hash"] = user.get_session_auth_hash()

        session.create()

        if expired:
            # Set expiry to the past
            session.set_expiry(-3600)  # Expired 1 hour ago
            session.save()

        return session

    return _create_session


@pytest.fixture
def websocket_scope():
    from webrockets import Connection

    def _create_scope(
        path="/ws/test/",
        headers=None,
        cookies=None,
        query_string="",
    ):
        return Connection(
            path=path,
            headers=headers or {},
            cookies=cookies or {},
            query_string=query_string,
        )

    return _create_scope


@pytest.fixture
def ws_server():
    yield TestWebsocketServer()



import importlib

if importlib.util.find_spec("django") is None:
    raise ImportError(
        'Django is required to use apps.py. Install with: pip install "webrockets[django]"'
    )

from django.apps import AppConfig


class WsRsAppConfig(AppConfig):
    name = "webrockets"
    verbose_name = "webrockets"

    def ready(self) -> None:
        from django.conf import settings

        broker_config = getattr(settings, "WEBSOCKET_BROKER", None)
        if broker_config:
            from webrockets import setup_broadcast

            setup_broadcast(broker_config)



import uvicorn
from fastapi import FastAPI, WebSocket, WebSocketDisconnect

app = FastAPI()


@app.websocket("/echo")
async def echo_endpoint(websocket: WebSocket):
    await websocket.accept()
    try:
        while True:
            data = await websocket.receive_bytes()
            await websocket.send_bytes(data)
    except WebSocketDisconnect:
        pass


if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=1234)



from threading import Thread
from time import sleep

from webrockets import WebsocketServer


class runserver:
    def __init__(self, server: WebsocketServer) -> None:
        self._server = server

    def __enter__(self):
        self._thread = Thread(target=self._server.start)
        self._thread.start()
        sleep(0.1)
        return self

    def __exit__(self, exc_type, exc, tb):
        self._server.stop()
        self._thread.join(timeout=5)
        if self._thread.is_alive():
            raise RuntimeError("WebSocket server did not shut down cleanly")



from chat import views
from django.contrib import admin
from django.urls import path

urlpatterns = [
    path("admin/", admin.site.urls),
    path("", views.index, name="index"),
    path("api/broadcast/", views.broadcast_message, name="broadcast"),
]



# /// script
# requires-python = ">=3.10"
# dependencies = [
#   "pydantic",
#   "webrockets"
# ]
# ///

import logging
from typing import Annotated, Any, Literal, cast

from pydantic import BaseModel, StringConstraints
from webrockets import Connection, IncomingConnection, Match, WebsocketServer
from webrockets.auth import AuthenticationFailed, BaseAuthentication

logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(name)s: %(message)s")


class ChatAuth(BaseAuthentication):
    def authenticate(self, conn: IncomingConnection) -> Any | None:
        name = conn.get_header("username")
        if name is None:
            raise AuthenticationFailed("username header not found", 4001)
        return name


class JoinMessage(BaseModel):
    type: Literal["join"]
    room: Annotated[str, StringConstraints(max_length=20)]


class JoinMessageResponse(BaseModel):
    type: Literal["join"] = "join"
    user: str


class ChatMessage(BaseModel):
    type: Literal["message"]
    msg: Annotated[str, StringConstraints(max_length=1024)]


class ChatMessageResponse(BaseModel):
    type: Literal["message"] = "message"
    user: str
    msg: Annotated[str, StringConstraints(max_length=1024)]


class LeaveMessageResponse(BaseModel):
    type: Literal["leave"] = "leave"
    user: str


server = WebsocketServer()

chat = server.create_route("chat", authentication_classes=[ChatAuth()])


@chat.connect("after")
def connect(conn: Connection):
    conn.send("Welcome!")


@chat.receive(match=Match("type", "join"), schema=JoinMessage)
def join(conn: Connection, data: JoinMessage):
    if len(conn.groups()) > 0:
        conn.send("You need to leave the current room to join another one")
        return

    conn.join(data.room)
    conn.broadcast([data.room], JoinMessageResponse(user=cast(str, conn.user)).model_dump_json())


@chat.receive(match=Match("type", "leave"))
def leave(conn: Connection, data: str | bytes):
    if len(conn.groups()) < 1:
        conn.send("You are not currently in a group")
        return

    group = conn.groups()[0]
    conn.send(f"Leaving {group}...")
    conn.leave(group)
    conn.broadcast([group], LeaveMessageResponse(user=cast(str, conn.user)).model_dump_json())


@chat.receive(match=Match("type", "message"), schema=ChatMessage)
def message(conn: Connection, data: ChatMessage):
    conn.broadcast(
        conn.groups(),
        ChatMessageResponse(user=cast(str, conn.user), msg=data.msg).model_dump_json(),
        exclude_self=True,
    )


@chat.disconnect
def disconnect(conn: Connection, code: int | None, reason: str | None):
    conn.broadcast(
        conn.groups(),
        LeaveMessageResponse(user=cast(str, conn.user)).model_dump_json(),
        exclude_self=True,
    )


try:
    server.start()
except KeyboardInterrupt:
    pass



import os
import sys

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")
sys.path.insert(0, os.path.dirname(__file__))

import django
from channels.generic.websocket import WebsocketConsumer
from channels.routing import ProtocolTypeRouter, URLRouter
from django.urls import path

django.setup()


class EchoConsumer(WebsocketConsumer):
    """Simple echo consumer."""

    def connect(self):
        self.accept()

    def disconnect(self, close_code):
        pass

    def receive(self, text_data=None, bytes_data=None):
        self.send(bytes_data=bytes_data)


application = ProtocolTypeRouter(
    {
        "websocket": URLRouter(
            [
                path("echo", EchoConsumer.as_asgi()),
            ]
        ),
    }
)

if __name__ == "__main__":
    from daphne.cli import CommandLineInterface

    sys.argv = ["daphne", "-b", "0.0.0.0", "-p", "8000", "server:application"]
    CommandLineInterface.entrypoint()



from typing import Any

from webrockets import IncomingConnection
from webrockets.auth import BaseAuthentication

class SessionAuthentication(BaseAuthentication):
    def __init__(self, session_cookie_name: str | None = None) -> None: ...
    @property
    def session_cookie_name(self) -> str: ...
    def authenticate(self, conn: IncomingConnection) -> Any | None: ...

class CookieTokenAuthentication(BaseAuthentication):
    cookie_name: str
    def authenticate(self, conn: IncomingConnection) -> Any | None: ...
    def validate_token(self, token: str) -> Any: ...

class HeaderTokenAuthentication(BaseAuthentication):
    header_name: str
    keyword: str
    def authenticate(self, conn: IncomingConnection) -> Any | None: ...
    def validate_token(self, token: str) -> Any: ...

class QueryStringTokenAuthentication(BaseAuthentication):
    query_param: str
    def authenticate(self, conn: IncomingConnection) -> Any | None: ...
    def validate_token(self, token: str) -> Any: ...



from abc import ABC, abstractmethod
from typing import Any

from webrockets import IncomingConnection

class AuthenticationFailed(Exception):
    detail: str
    close_code: int
    def __init__(self, detail: str = "Authentication failed", close_code: int = 4003) -> None: ...

class BaseAuthentication(ABC):
    @abstractmethod
    def authenticate(self, conn: IncomingConnection) -> Any | None: ...



from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent

SECRET_KEY = "benchmark-secret-key-not-for-production"
DEBUG = False
ALLOWED_HOSTS = ["*"]

INSTALLED_APPS = [
    "daphne",
    "django.contrib.contenttypes",
    "channels",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "/tmp/bench_channels.db",
    }
}

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

ASGI_APPLICATION = "server.application"



import importlib.util

from webrockets import WebsocketServer

if importlib.util.find_spec("django") is None:
    raise ImportError(
        'Django is required for webrockets.django. Install with: pip install "webrockets[django]"'
    )

from django.conf import settings

from .auth import (
    AuthenticationFailed,
    CookieTokenAuthentication,
    HeaderTokenAuthentication,
    QueryStringTokenAuthentication,
    SessionAuthentication,
)

server = WebsocketServer(
    getattr(settings, "WEBSOCKET_HOST", "0.0.0.0"),
    getattr(settings, "WEBSOCKET_PORT", 46290),
    getattr(settings, "WEBSOCKET_BROKER", None),
)

__all__ = [
    "AuthenticationFailed",
    "SessionAuthentication",
    "CookieTokenAuthentication",
    "HeaderTokenAuthentication",
    "QueryStringTokenAuthentication",
    "server",
]



"""
WebSocket routes for the chat application.

This module is automatically discovered by webrockets when running:
    python manage.py runwebsockets
"""

import json
import logging

from webrockets import Match, broadcast
from webrockets.django import server

logger = logging.getLogger(__name__)


# Create the chat route (no authentication for this example)
chat = server.create_route("ws/chat/")


@chat.connect("after")
def on_connect(conn):
    logger.info(f"Client connected: {conn.path}")
    conn.send(
        json.dumps(
            {
                "type": "system",
                "message": "Connected to chat server. Send a 'join' message to enter a room.",
            }
        )
    )


@chat.receive(match=Match("type", "join"))
def handle_join(conn, msg: str | bytes):
    msg = json.loads(msg)
    room = msg.get("room", "general")
    username = msg.get("username", "Anonymous")

    # Store username on connection
    conn.user = username

    # Leave any existing rooms
    for group in conn.groups():
        conn.leave(group)

    # Join the new room
    conn.join(room)

    logger.info(f"{username} joined room: {room}")

    # Notify the user
    conn.send(
        json.dumps(
            {
                "type": "joined",
                "room": room,
                "message": f"You joined {room}",
            }
        )
    )

    # Broadcast to room that user joined (via Redis for multi-server)
    broadcast(
        [room],
        json.dumps(
            {
                "type": "user_joined",
                "username": username,
                "room": room,
            }
        ),
    )


@chat.receive(match=Match("type", "leave"))
def handle_leave(conn, msg):
    username = getattr(conn, "user", "Anonymous")
    groups = conn.groups()

    if not groups:
        conn.send(
            json.dumps(
                {
                    "type": "error",
                    "message": "You are not in any room",
                }
            )
        )
        return

    room = groups[0]
    conn.leave(room)

    logger.info(f"{username} left room: {room}")

    # Notify the user
    conn.send(
        json.dumps(
            {
                "type": "left",
                "room": room,
                "message": f"You left {room}",
            }
        )
    )

    # Broadcast to room that user left (via Redis for multi-server)
    broadcast(
        [room],
        json.dumps(
            {
                "type": "user_left",
                "username": username,
                "room": room,
            }
        ),
    )


@chat.receive(match=Match("type", "message"))
def handle_message(conn, msg):
    username = getattr(conn, "user", "Anonymous")
    msg = json.loads(msg)
    content = msg.get("message", "")
    groups = conn.groups()

    if not groups:
        conn.send(
            json.dumps(
                {
                    "type": "error",
                    "message": "Join a room first to send messages",
                }
            )
        )
        return

    if not content:
        return

    room = groups[0]

    logger.info(f"{username} in {room}: {content}")

    # Broadcast message to room (via Redis for multi-server)
    broadcast(
        [room],
        json.dumps(
            {
                "type": "message",
                "sender": username,
                "message": content,
                "room": room,
            }
        ),
    )


@chat.receive
def fallback(conn, data):
    conn.send(
        json.dumps(
            {
                "type": "error",
                "message": "Unknown message type",
            }
        )
    )


@chat.disconnect
def on_disconnect(conn, code, reason):
    username = getattr(conn, "user", "Anonymous")
    groups = conn.groups()

    logger.info(f"Client disconnected: {username} (code={code})")

    for room in groups:
        broadcast(
            [room],
            json.dumps(
                {
                    "type": "user_left",
                    "username": username,
                    "room": room,
                }
            ),
        )



import asyncio
import json

import pytest
from pydantic import BaseModel
from webrockets import Connection, IncomingConnection, Match, WebsocketServer
from webrockets.client import ConnectionClosed, aconnect
from webrockets.test import runserver


class TestCallbacks:
    @pytest.mark.asyncio
    async def test_connect(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.connect("before")
        def before_connect(conn: IncomingConnection):
            assert isinstance(conn, IncomingConnection)
            assert conn.path == "/ws"

        @route.connect("after")
        def after_connect(conn: Connection):
            assert isinstance(conn, Connection)
            assert conn.path == "/ws"
            conn.close(code=1000)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                try:
                    await asyncio.wait_for(ws.recv(), timeout=2.0)
                    pytest.fail("Expected connection to close")
                except ConnectionClosed as e:
                    assert e.code == 1000

    @pytest.mark.asyncio
    async def test_async_connect(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.connect("before")
        async def before_connect(conn: IncomingConnection):
            assert isinstance(conn, IncomingConnection)
            assert conn.path == "/ws"

        @route.connect("after")
        async def after_connect(conn: Connection):
            assert isinstance(conn, Connection)
            assert conn.path == "/ws"
            await conn.aclose(code=1000)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                try:
                    await asyncio.wait_for(ws.recv(), timeout=2.0)
                    pytest.fail("Expected connection to close")
                except ConnectionClosed as e:
                    assert e.code == 1000

    @pytest.mark.asyncio
    async def test_receive(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.receive
        def receive(conn: Connection, data: str | bytes):
            assert isinstance(conn, Connection)
            assert data == "hi"
            conn.close(code=1000)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                try:
                    await ws.send("hi")
                    await asyncio.wait_for(ws.recv(), timeout=2.0)
                    pytest.fail("Expected connection to close")
                except ConnectionClosed as e:
                    assert e.code == 1000

    @pytest.mark.asyncio
    async def test_async_receive(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.receive
        async def receive(conn: Connection, data: str | bytes):
            assert isinstance(conn, Connection)
            assert data == "hi"
            await conn.aclose(code=1000)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                try:
                    await ws.send("hi")
                    await asyncio.wait_for(ws.recv(), timeout=2.0)
                    pytest.fail("Expected connection to close")
                except ConnectionClosed as e:
                    assert e.code == 1000

    @pytest.mark.asyncio
    async def test_disconnect(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.disconnect
        def disconnect(conn: Connection, code: int | None, reason: str | None):
            assert code == 1000
            assert reason == "test"

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.close(code=1000, reason="test")

    @pytest.mark.asyncio
    async def test_async_disconnect(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.disconnect
        async def disconnect(conn: Connection, code: int | None, reason: str | None):
            assert code == 1000
            assert reason == "test"

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.close(code=1000, reason="test")


class TestMessageSending:
    @pytest.mark.asyncio
    async def test_send(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.receive
        def receive(conn: Connection, data: str | bytes):
            conn.send(data)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.send("hi")
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == "hi"

    @pytest.mark.asyncio
    async def test_async_send(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.receive
        async def receive(conn: Connection, data: str | bytes):
            await conn.asend(data)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.send("hi")
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == "hi"

    @pytest.mark.asyncio
    async def test_send_and_receive_binary(self, ws_server):
        route = ws_server.create_route("ws")

        @route.receive
        async def receive(conn: Connection, data: str | bytes):
            await conn.asend(data)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                binary_data = b"\x00\x01\x02\x03\xff\xfe"
                await ws.send(binary_data)
                response = await asyncio.wait_for(ws.recv(), timeout=2.0)

                assert response == binary_data
                assert isinstance(response, bytes)

    @pytest.mark.asyncio
    async def test_binary_preserves_content(self, ws_server):
        route = ws_server.create_route("ws")

        @route.receive
        async def receive(conn: Connection, data: str | bytes):
            await conn.asend(data)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                test_data = [
                    b"\x00" * 100,  # Null bytes
                    b"\xff" * 100,  # High bytes
                    bytes(range(256)),  # All byte values
                    b"mixed\x00binary\xffdata",  # Mixed content
                ]

                for data in test_data:
                    await ws.send(data)
                    response = await asyncio.wait_for(ws.recv(), timeout=2.0)
                    assert response == data
                    assert isinstance(response, bytes)


class TestGroups:
    @pytest.mark.asyncio
    async def test_join(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.connect("after")
        def connect(conn: Connection):
            conn.join("chat")

        @route.receive
        def receive(conn: Connection, _):
            conn.send(json.dumps(conn.groups()))

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.send("hi")
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == '["chat"]'

    @pytest.mark.asyncio
    async def test_leave(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws", "chat")

        @route.connect("after")
        def connect(conn: Connection):
            conn.join("test")
            conn.leave("chat")

        @route.receive
        def receive(conn: Connection, _):
            conn.send(json.dumps(conn.groups()))

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.send("hi")
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == '["test"]'


class TestBroadcast:
    @pytest.mark.asyncio
    async def test_broadcast(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.connect("after")
        def connect(conn: Connection):
            conn.join("chat")

        @route.receive
        def receive(conn: Connection, data: str | bytes):
            conn.broadcast(["chat"], data)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                async with aconnect(f"ws://{ws_server.addr()}/ws") as ws2:
                    await ws.send("hi")

                    msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                    assert msg == "hi"

                    msg = await asyncio.wait_for(ws2.recv(), timeout=2.0)
                    assert msg == "hi"

    @pytest.mark.asyncio
    async def test_async_broadcast(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.connect("after")
        def connect(conn: Connection):
            conn.join("chat")

        @route.receive
        async def receive(conn: Connection, data: str | bytes):
            await conn.abroadcast(conn.groups(), data)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                async with aconnect(f"ws://{ws_server.addr()}/ws") as ws2:
                    await ws.send("hi")

                    msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                    assert msg == "hi"

                    msg = await asyncio.wait_for(ws2.recv(), timeout=2.0)
                    assert msg == "hi"


class TestPatternMatching:
    @pytest.mark.asyncio
    async def test_simple_pattern_match(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.receive(match=Match("type", "message"))
        async def receive(conn: Connection, data: str | bytes):
            await conn.asend(data)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.send("hi")

                payload = json.dumps({"type": "message", "data": "hi"})
                await ws.send(payload)
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == payload

    @pytest.mark.asyncio
    async def test_match_on_multiple_keys(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.receive(match=Match(["type", "action"], "message"))
        async def receive(conn: Connection, data: str | bytes):
            await conn.asend(data)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.send("hi")

                payload = json.dumps({"type": "message", "data": "hi"})
                await ws.send(payload)
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == payload

                payload = json.dumps({"action": "message", "data": "hi"})
                await ws.send(payload)
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == payload

    @pytest.mark.asyncio
    async def test_match_on_multiple_values(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.receive(match=Match("type", ["message", "mention"]))
        async def receive(conn: Connection, data: str | bytes):
            await conn.asend(data)

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.send("hi")

                payload = json.dumps({"type": "message", "data": "hi"})
                await ws.send(payload)
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == payload

                payload = json.dumps({"type": "mention", "data": "hi"})
                await ws.send(payload)
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == payload

    @pytest.mark.asyncio
    async def test_generic_fallback(self, ws_server: WebsocketServer):
        route = ws_server.create_route("ws")

        @route.receive(match=Match("type", "message"))
        async def receive_message(conn: Connection, data: str | bytes):
            await conn.asend(data)

        @route.receive
        async def receive(conn: Connection, data: str | bytes):
            await conn.asend("generic")

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                await ws.send("hi")
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == "generic"

    @pytest.mark.asyncio
    async def test_match_remove_key(self, ws_server: WebsocketServer):
        class ChatPayload(BaseModel):
            content: str

        route = ws_server.create_route("ws")

        @route.receive(match=Match("type", "message", remove_key=True), schema=ChatPayload)
        async def receive_message(conn: Connection, data: ChatPayload):
            await conn.asend(data.model_dump_json())

        with runserver(ws_server):
            async with aconnect(f"ws://{ws_server.addr()}/ws") as ws:
                payload = json.dumps({"type": "message", "content": "hi"})
                await ws.send(payload)
                msg = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert msg == '{"content":"hi"}'



from threading import Thread
from time import sleep

from django.test import TestCase
from django.utils.module_loading import autodiscover_modules

from webrockets.django import server


class WebsocketTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        autodiscover_modules("websockets", "sockets", "views")

    def setUp(self):
        super().setUp()
        self._websocket_thread = Thread(target=server.start)
        self._websocket_thread.start()
        sleep(0.1)

    def tearDown(self):
        super().tearDown()
        server.stop()
        self._websocket_thread.join(timeout=5)
        if self._websocket_thread.is_alive():
            raise RuntimeError("WebSocket server did not shut down cleanly")



from types import TracebackType
from typing import Self

class ClientConfig:
    extra_headers: dict[str, str]

    def __init__(self, extra_headers: dict[str, str] | None = None) -> None: ...

class Client:
    config: ClientConfig

    def __init__(self, config: ClientConfig | None = None) -> None: ...
    def connect(self, url: str) -> None: ...
    def send(self, data: str | bytes) -> None: ...
    def recv(self) -> str | bytes: ...
    def close(self, code=1000, reason="") -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None: ...

def connect(url: str, config: ClientConfig | None = None) -> Client: ...

class AsyncClient:
    config: ClientConfig

    def __init__(self, config: ClientConfig | None = None) -> None: ...
    async def connect(self, url: str) -> None: ...
    async def send(self, data: str | bytes) -> None: ...
    async def recv(self) -> str | bytes: ...
    async def close(self, code=1000, reason="") -> None: ...
    async def __aenter__(self) -> Self: ...
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None: ...

def aconnect(url: str, config: ClientConfig | None = None) -> AsyncClient: ...



# /// script
# requires-python = ">=3.10"
# dependencies = [
#   "websockets",
#   "rich",
# ]
# ///

import asyncio
import json
import sys
import termios
import tty
from datetime import datetime
from queue import Empty, Queue
from threading import Thread

import websockets
from rich.console import Console
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.text import Text


class ChatClient:
    def __init__(self, name: str, room: str):
        self.name = name
        self.room = room
        self.console = Console()
        self.messages: list[tuple[str, str, str]] = []
        self.input_buffer = ""
        self.running = True
        self.ws: websockets.ClientConnection | None = None
        self.input_queue: Queue[str | None] = Queue()

    def render_messages(self) -> Panel:
        text = Text()
        for timestamp, sender, message in self.messages[-20:]:
            text.append(f"[{timestamp}] ", style="dim")
            if sender == "system":
                text.append(f"{message}\n", style="yellow italic")
            elif sender == self.name:
                text.append(f"{sender}: ", style="bold green")
                text.append(f"{message}\n")
            else:
                text.append(f"{sender}: ", style="bold cyan")
                text.append(f"{message}\n")
        if not self.messages:
            text.append("No messages yet...", style="dim italic")
        return Panel(text, title=f"Room: {self.room}", border_style="blue")

    def render_input(self) -> Panel:
        return Panel(
            Text(f"> {self.input_buffer}█", style="white"),
            title="Message (Enter to send, Ctrl+C to quit)",
            border_style="green",
            height=3,
        )

    def render(self) -> Layout:
        layout = Layout()
        layout.split_column(
            Layout(name="messages", ratio=9),
            Layout(name="input", ratio=1, minimum_size=3),
        )
        layout["messages"].update(self.render_messages())
        layout["input"].update(self.render_input())
        return layout

    def add_message(self, sender: str, message: str):
        timestamp = datetime.now().strftime("%H:%M:%S")
        self.messages.append((timestamp, sender, message))

    def input_thread(self):
        """Read input in a separate thread to avoid blocking."""
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setcbreak(fd)  # Use cbreak instead of raw - allows Ctrl+C
            while self.running:
                char = sys.stdin.read(1)
                if char == "\x03":  # Ctrl+C
                    self.input_queue.put(None)
                    break
                self.input_queue.put(char)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    async def process_input(self):
        """Process input from the queue."""
        while self.running:
            try:
                char = self.input_queue.get_nowait()
                if char is None:  # Ctrl+C signal
                    self.running = False
                    break
                if char in ("\r", "\n"):
                    if self.input_buffer.strip():
                        msg = self.input_buffer.strip()
                        if self.ws:
                            await self.ws.send(json.dumps({"type": "message", "msg": msg}))
                            self.add_message(self.name, msg)
                    self.input_buffer = ""
                elif char == "\x7f" or char == "\x08":  # Backspace
                    self.input_buffer = self.input_buffer[:-1]
                elif char.isprintable():
                    self.input_buffer += char
            except Empty:
                await asyncio.sleep(0.02)

    async def recv_messages(self):
        """Receive messages from websocket."""
        assert self.ws is not None
        try:
            async for raw in self.ws:
                try:
                    data = json.loads(raw)
                    msg_type = data.get("type")
                    if msg_type == "message":
                        sender = data.get("user", "unknown")
                        self.add_message(sender, data.get("msg", ""))
                    elif msg_type == "join":
                        self.add_message("system", f"{data.get('user', 'Someone')} joined")
                    elif msg_type == "leave":
                        self.add_message("system", f"{data.get('user', 'Someone')} left")
                    else:
                        self.add_message("system", str(data))
                except json.JSONDecodeError:
                    self.add_message("system", str(raw))
        except websockets.ConnectionClosed:
            self.add_message("system", "Connection closed")
            self.running = False

    async def run(self):
        # Start input thread
        input_thread = Thread(target=self.input_thread, daemon=True)
        input_thread.start()

        try:
            async with websockets.connect(
                "ws://localhost:46290/chat",
                additional_headers={"username": self.name},
            ) as ws:
                self.ws = ws
                await ws.send(json.dumps({"type": "join", "room": self.room}))
                self.add_message("system", f"Connected to room '{self.room}'")

                recv_task = asyncio.create_task(self.recv_messages())
                input_task = asyncio.create_task(self.process_input())

                with Live(
                    self.render(),
                    console=self.console,
                    refresh_per_second=30,
                    screen=True,
                ) as live:
                    while self.running:
                        live.update(self.render())
                        await asyncio.sleep(0.033)  # ~30fps

                recv_task.cancel()
                input_task.cancel()

        except ConnectionRefusedError:
            self.console.print("[red]Could not connect to server. Is it running?[/red]")
        except Exception as e:
            self.console.print(f"[red]Error: {e}[/red]")


def main():
    console = Console()
    console.print("[bold blue]Chat Client[/bold blue]")

    name = console.input("[green]What's your name?[/green] ")
    room = console.input("[green]Room name?[/green] ")

    if not name.strip() or not room.strip():
        console.print("[red]Name and room are required[/red]")
        return

    client = ChatClient(name.strip(), room.strip())

    try:
        asyncio.run(client.run())
    except KeyboardInterrupt:
        pass

    console.print("\n[yellow]Goodbye![/yellow]")


if __name__ == "__main__":
    main()



import json

from django.http import JsonResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from webrockets import broadcast


def index(request):
    """Render the chat interface."""
    return render(request, "index.html")


@csrf_exempt
@require_POST
def broadcast_message(request):
    """
    API endpoint to broadcast a message from Django.

    This demonstrates how to send messages from Django to WebSocket clients
    using the broadcast() function and Redis broker.
    """
    try:
        data = json.loads(request.body)
        room = data.get("room", "general")
        message = data.get("message", "")
        sender = data.get("sender", "Server")

        if not message:
            return JsonResponse({"error": "Message is required"}, status=400)

        # Broadcast to all clients in the room via Redis
        broadcast(
            [room],
            json.dumps(
                {
                    "type": "message",
                    "sender": sender,
                    "message": message,
                    "from_api": True,
                }
            ),
        )

        return JsonResponse({"status": "sent", "room": room})

    except json.JSONDecodeError:
        return JsonResponse({"error": "Invalid JSON"}, status=400)



import asyncio
import json

import pytest
from testcontainers.rabbitmq import RabbitMqContainer
from testcontainers.redis import RedisContainer
from webrockets import (
    BrokerConfig,
    Connection,
    WebsocketServer,
    abroadcast,
    broadcast,
    reset_broadcast,
    setup_broadcast,
)
from webrockets.client import aconnect
from webrockets.test import runserver


@pytest.fixture(scope="module")
def redis_container():
    with RedisContainer("redis:7-alpine") as redis:
        yield redis


@pytest.fixture
def broker_config(redis_container) -> BrokerConfig:
    redis_url = f"redis://{redis_container.get_container_host_ip()}:{redis_container.get_exposed_port(6379)}"
    config: BrokerConfig = {
        "type": "redis",
        "url": redis_url,
        "channel": "ws_test_broadcast",
    }
    setup_broadcast(config)
    return config


@pytest.fixture
def redis_ws_server(broker_config: BrokerConfig) -> WebsocketServer:
    server = WebsocketServer(broker=broker_config)
    route = server.create_route("ws/redis-test/", "redis-test")

    @route.receive
    def on_receive(conn: Connection, data: str | bytes):
        conn.send(f"redis-ack: {data}")

    return server


class TestRedisBroker:
    @pytest.mark.asyncio
    async def test_redis_broadcaster_send(self, redis_ws_server: WebsocketServer):
        with runserver(redis_ws_server):
            async with aconnect(f"ws://{redis_ws_server.addr()}/ws/redis-test/") as ws:
                broadcast(
                    groups=["redis-test"],
                    message=json.dumps({"type": "test", "data": "hello from redis"}),
                )

                response = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert "hello from redis" in str(response)

    @pytest.mark.asyncio
    async def test_redis_broadcaster_asend(self, redis_ws_server: WebsocketServer):
        with runserver(redis_ws_server):
            async with aconnect(f"ws://{redis_ws_server.addr()}/ws/redis-test/") as ws:
                await abroadcast(
                    groups=["redis-test"],
                    message=json.dumps({"type": "async_test", "data": "async hello"}),
                )

                response = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert "async hello" in str(response)

    @pytest.mark.asyncio
    async def test_redis_multiple_broadcasts(self, redis_ws_server: WebsocketServer):
        with runserver(redis_ws_server):
            async with aconnect(f"ws://{redis_ws_server.addr()}/ws/redis-test/") as ws:
                messages = ["first", "second", "third"]
                for msg in messages:
                    broadcast(groups=["redis-test"], message=msg)

                received = []
                for _ in range(len(messages)):
                    try:
                        response = await asyncio.wait_for(ws.recv(), timeout=2.0)
                        received.append(response)
                    except asyncio.TimeoutError:
                        break

                assert len(received) == len(messages)
                for msg in messages:
                    assert any(msg in r for r in received)

    @pytest.mark.asyncio
    async def test_redis_broadcast_to_multiple_clients(self, redis_ws_server: WebsocketServer):
        with runserver(redis_ws_server):
            async with aconnect(f"ws://{redis_ws_server.addr()}/ws/redis-test/") as ws1:
                async with aconnect(f"ws://{redis_ws_server.addr()}/ws/redis-test/") as ws2:
                    broadcast(groups=["redis-test"], message="broadcast to all")

                    r1 = await asyncio.wait_for(ws1.recv(), timeout=2.0)
                    r2 = await asyncio.wait_for(ws2.recv(), timeout=2.0)

                    assert "broadcast to all" in str(r1)
                    assert "broadcast to all" in str(r2)

    @pytest.mark.asyncio
    async def test_redis_broadcast_works_from_handler(self, redis_ws_server: WebsocketServer):
        route = redis_ws_server.create_route("ws/redis-test-broadcast/", "redis-test")

        @route.receive
        def on_receive(conn: Connection, data: str | bytes):
            broadcast(groups=["redis-test"], message="broadcast to all")

        with runserver(redis_ws_server):
            async with aconnect(f"ws://{redis_ws_server.addr()}/ws/redis-test-broadcast/") as ws:
                await ws.send("hi")
                r = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert "broadcast to all" in str(r)


@pytest.fixture(scope="module")
def amqp_container():
    with RabbitMqContainer("rabbitmq:3-alpine") as rabbitmq:
        yield rabbitmq


@pytest.fixture
def amqp_broker_config(amqp_container) -> BrokerConfig:
    amqp_url = f"amqp://guest:guest@{amqp_container.get_container_host_ip()}:{amqp_container.get_exposed_port(5672)}"
    config: BrokerConfig = {
        "type": "amqp",
        "url": amqp_url,
        "exchange": "ws_test_broadcast",
    }
    reset_broadcast()
    setup_broadcast(config)
    return config


@pytest.fixture
def amqp_ws_server(amqp_broker_config: BrokerConfig) -> WebsocketServer:
    server = WebsocketServer(broker=amqp_broker_config)
    route = server.create_route("ws/amqp-test/", "amqp-test")

    @route.receive
    def on_receive(conn: Connection, data: str | bytes):
        conn.send(f"amqp-ack: {data}")

    return server


class TestAmqpBroker:
    @pytest.mark.asyncio
    async def test_amqp_broadcaster_send(self, amqp_ws_server: WebsocketServer):
        with runserver(amqp_ws_server):
            async with aconnect(f"ws://{amqp_ws_server.addr()}/ws/amqp-test/") as ws:
                broadcast(
                    groups=["amqp-test"],
                    message=json.dumps({"type": "test", "data": "hello from amqp"}),
                )

                response = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert "hello from amqp" in str(response)

    @pytest.mark.asyncio
    async def test_amqp_broadcaster_asend(self, amqp_ws_server: WebsocketServer):
        with runserver(amqp_ws_server):
            async with aconnect(f"ws://{amqp_ws_server.addr()}/ws/amqp-test/") as ws:
                await abroadcast(
                    groups=["amqp-test"],
                    message=json.dumps({"type": "async_test", "data": "async hello"}),
                )

                response = await asyncio.wait_for(ws.recv(), timeout=2.0)
                assert "async hello" in str(response)

    @pytest.mark.asyncio
    async def test_amqp_multiple_broadcasts(self, amqp_ws_server: WebsocketServer):
        with runserver(amqp_ws_server):
            async with aconnect(f"ws://{amqp_ws_server.addr()}/ws/amqp-test/") as ws:
                messages = ["first", "second", "third"]
                for msg in messages:
                    broadcast(groups=["amqp-test"], message=msg)

                received = []
                for _ in range(len(messages)):
                    try:
                        response = await asyncio.wait_for(ws.recv(), timeout=2.0)
                        received.append(response)
                    except asyncio.TimeoutError:
                        break

                assert len(received) == len(messages)
                for msg in messages:
                    assert any(msg in r for r in received)

    @pytest.mark.asyncio
    async def test_amqp_broadcast_to_multiple_clients(self, amqp_ws_server: WebsocketServer):
        with runserver(amqp_ws_server):
            async with aconnect(f"ws://{amqp_ws_server.addr()}/ws/amqp-test/") as ws1:
                async with aconnect(f"ws://{amqp_ws_server.addr()}/ws/amqp-test/") as ws2:
                    broadcast(groups=["amqp-test"], message="broadcast to all")

                    r1 = await asyncio.wait_for(ws1.recv(), timeout=2.0)
                    r2 = await asyncio.wait_for(ws2.recv(), timeout=2.0)

                    assert "broadcast to all" in str(r1)
                    assert "broadcast to all" in str(r2)


class TestBroadcasterErrors:
    def test_broadcaster_invalid_type(self):
        with pytest.raises(ValueError, match="unknown broker type"):
            setup_broadcast({"type": "invalid"})



from collections.abc import Callable, Coroutine, Iterable
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    TypedDict,
    TypeVar,
    overload,
)

from webrockets.auth import BaseAuthentication

if TYPE_CHECKING:
    from pydantic import BaseModel

# Type aliases for Match parameters
MatchKey = str | Iterable[str]
MatchValue = str | int | Literal["*"] | Iterable[str | int | Literal["*"]]

class _RedisBrokerConfigOptional(TypedDict, total=False):
    url: str  # default: "redis://localhost:6379"
    channel: str  # default: "ws_broadcast"

class RedisBrokerConfig(_RedisBrokerConfigOptional):
    type: Literal["redis"]

class _AmqpBrokerConfigOptional(TypedDict, total=False):
    url: str  # default: "amqp://localhost:5672"
    exchange: str  # default: "ws_broadcast"
    queue: str | None  # default: auto-generated UUID
    routing_key: str  # default: "#"

class AmqpBrokerConfig(_AmqpBrokerConfigOptional):
    type: Literal["amqp"]

BrokerConfig = RedisBrokerConfig | AmqpBrokerConfig

class BaseConnection:
    path: str
    query_string: str
    headers: dict[str, str]
    cookies: dict[str, str]
    user: Any | None

    def __init__(
        self,
        path: str,
        query_string: str,
        headers: dict[str, str],
        cookies: dict[str, str],
    ) -> None: ...
    def get_cookie(self, name: str) -> str | None: ...
    def get_header(self, name: str) -> str | None: ...

class IncomingConnection(BaseConnection): ...

class Connection(BaseConnection):
    def send(self, msg: str | bytes) -> None: ...
    async def asend(self, msg: str | bytes) -> None: ...
    def close(self, code: int = 1000, reason: str = "") -> None: ...
    async def aclose(self, code: int = 1000, reason: str = "") -> None: ...
    def broadcast(
        self, groups: list[str], msg: str | bytes, exclude_self: bool = False
    ) -> None: ...
    async def abroadcast(
        self, groups: list[str], msg: str | bytes, exclude_self: bool = False
    ) -> None: ...
    def join(self, group: str) -> bool:
        """Join a group dynamically. Returns True if newly joined, False if already a member."""
        ...
    def leave(self, group: str) -> bool:
        """Leave a group dynamically. Returns True if was a member, False otherwise."""
        ...
    def groups(self) -> list[str]:
        """Get all groups this connection belongs to."""
        ...
    def group_size(self, group: str) -> int:
        """Get the number of connections in a group."""
        ...

T_Connection = TypeVar("T_Connection", bound=IncomingConnection | Connection)
T_Schema = TypeVar("T_Schema", bound="BaseModel" | str)

class ConnectDecorator(Generic[T_Connection]):
    def __call__(
        self, func: Callable[[T_Connection], None | Coroutine[Any, Any, None]]
    ) -> Callable[[T_Connection], None | Coroutine[Any, Any, None]]: ...

class ReceiveDecorator(Generic[T_Schema]):
    def __call__(
        self, func: Callable[[Connection, T_Schema], None | Coroutine[Any, Any, None]]
    ) -> Callable[[Connection, T_Schema], None | Coroutine[Any, Any, None]]: ...

class Match:
    """
    Pattern matcher for receive handlers.

    Supports matching on JSON message fields with:
    - Single or multiple keys: Match("type", ...) or Match(["type", "action"], ...)
    - Single or multiple values: Match(..., "message") or Match(..., ["msg", "notify"])
    - Integer values: Match("code", 1) or Match("code", [1, 2, 3])
    - Wildcard matching: Match("type", "*") matches any value
    - Mixed types: Match("type", ["message", 1, "*"])

    Args:
        key: Field name(s) to match on (string or iterable of strings)
        value: Value(s) to match (string, int, "*" for wildcard, or iterable)
        remove_key: If True, removes the matched key from the JSON before passing to handler
    """

    key: list[str]
    value: list[str | int]
    remove_key: bool

    def __init__(
        self,
        key: MatchKey,
        value: MatchValue,
        *,
        remove_key: bool = False,
    ) -> None: ...

class WebsocketRoute:
    path: str
    default_group: str | None
    authentication_classes: list[BaseAuthentication] | None = None

    @overload
    def connect(
        self,
        when: Literal["before"],
    ) -> ConnectDecorator[IncomingConnection]: ...
    @overload
    def connect(
        self,
        when: Literal["after"],
    ) -> ConnectDecorator[Connection]: ...
    @overload
    def receive(
        self,
        func: Callable[[Connection, str | bytes], None | Coroutine[Any, Any, None]],
        /,
    ) -> Callable[[Connection, str | bytes], None | Coroutine[Any, Any, None]]: ...
    @overload
    def receive(
        self,
        /,
        match: Match,
    ) -> ReceiveDecorator[str]: ...
    @overload
    def receive(
        self,
        /,
        match: Match,
        schema: type[T_Schema],
    ) -> ReceiveDecorator[T_Schema]: ...
    def disconnect(
        self,
        func: Callable[[Connection, int | None, str | None], None | Coroutine[Any, Any, None]],
    ) -> Callable[[Connection, int | None, str | None], None | Coroutine[Any, Any, None]]: ...

class WebsocketServer:
    def __init__(
        self,
        host: str = "0.0.0.0",
        port: int = 46290,
        broker: BrokerConfig | None = None,
    ) -> None: ...
    def start(self) -> None: ...
    def stop(self) -> None: ...
    def create_route(
        self,
        path: str,
        default_group: str | None = None,
        authentication_classes: list[BaseAuthentication] | None = None,
    ) -> WebsocketRoute: ...
    def addr(self) -> str: ...

# Broker functions for external broadcasting (e.g., from Django views, Celery tasks)
def setup_broadcast(config: BrokerConfig) -> None:
    """
    Initialize the broadcaster with a broker configuration.
    Must be called before using broadcast() or abroadcast().
    """
    ...

def reset_broadcast() -> None:
    """
    Reset the broadcaster, allowing setup_broadcast() to be called again.
    Primarily useful for testing when switching between broker configurations.
    """
    ...

def broadcast(groups: list[str], message: str) -> None:
    """
    Publish a message to the specified groups via the configured broker.
    This is a blocking call that publishes synchronously.

    Args:
        groups: List of group names to broadcast to
        message: The message payload (typically JSON string)
    """
    ...

async def abroadcast(groups: list[str], message: str) -> None:
    """
    Async version of broadcast().
    Publish a message to the specified groups via the configured broker.

    Args:
        groups: List of group names to broadcast to
        message: The message payload (typically JSON string)
    """
    ...



import pytest
from webrockets import Connection
from webrockets.auth import BaseAuthentication


class TestSocketViewCreation:
    def test_create_view_basic(self, ws_server):
        view = ws_server.create_route("ws/test/", "test_group")

        assert view.path == "ws/test/"
        assert view.default_group == "test_group"

    def test_create_view_without_default_group(self, ws_server):
        view = ws_server.create_route("ws/no_group/")

        assert view.path == "ws/no_group/"
        assert view.default_group is None

    def test_create_view_with_auth_classes(self, ws_server):
        class MockAuth(BaseAuthentication):
            def authenticate(self, conn):
                return None

        view = ws_server.create_route(
            "ws/secure/",
            "secure_group",
            authentication_classes=[MockAuth()],
        )

        assert view.path == "ws/secure/"
        assert view.default_group == "secure_group"

    def test_create_multiple_views(self, ws_server):
        view1 = ws_server.create_route("ws/chat/", "chat")
        view2 = ws_server.create_route("ws/notifications/", "notifications")

        assert view1.path == "ws/chat/"
        assert view2.path == "ws/notifications/"
        assert view1.default_group == "chat"
        assert view2.default_group == "notifications"


class TestConnectDecorator:
    def test_connect_before_decorator_registers_callback(self, ws_server):
        view = ws_server.create_route("ws/connect_test/", "connect_group")
        callback_called = []

        @view.connect("before")
        def on_connect(conn):
            callback_called.append(conn)

        assert callable(on_connect)

    def test_connect_after_decorator_registers_callback(self, ws_server):
        view = ws_server.create_route("ws/connect_after_test/", "connect_after_group")
        callback_called = []

        @view.connect("after")
        def on_connect(conn):
            callback_called.append(conn)

        assert callable(on_connect)

    def test_connect_decorator_preserves_function(self, ws_server):
        view = ws_server.create_route("ws/connect_preserve/", "connect_preserve")

        @view.connect("before")
        def my_connect_handler(conn):
            return "connected"

        conn = Connection("/ws/", "", {}, {})
        result = my_connect_handler(conn)
        assert result == "connected"

    def test_connect_invalid_argument_raises_error(self, ws_server):
        view = ws_server.create_route("ws/connect_invalid/", "connect_invalid")

        with pytest.raises(ValueError, match="'before' or 'after'"):
            view.connect("invalid")


class TestReceiveDecorator:
    def test_receive_decorator_registers_callback(self, ws_server):
        view = ws_server.create_route("ws/receive_test/", "receive_group")

        @view.receive
        def on_receive(conn, data):
            pass

        assert callable(on_receive)

    def test_receive_decorator_preserves_function(self, ws_server):
        view = ws_server.create_route("ws/receive_preserve/", "receive_preserve")

        @view.receive
        def my_receive_handler(conn, data):
            return f"Received: {data}"

        conn = Connection("/ws/", "", {}, {})
        result = my_receive_handler(conn, "hello")
        assert result == "Received: hello"


class TestDisconnectDecorator:
    def test_disconnect_decorator_registers_callback(self, ws_server):
        view = ws_server.create_route("ws/disconnect_test/", "disconnect_group")

        @view.disconnect
        def on_disconnect(conn, code=None, reason=None):
            pass

        assert callable(on_disconnect)

    def test_disconnect_decorator_preserves_function(self, ws_server):
        view = ws_server.create_route("ws/disconnect_preserve/", "disconnect_preserve")

        @view.disconnect
        def my_disconnect_handler(conn, code=None, reason=None):
            return f"Disconnected: {code}"

        conn = Connection("/ws/", "", {}, {})
        result = my_disconnect_handler(conn, 1000, "normal")
        assert result == "Disconnected: 1000"


class TestFullViewSetup:
    def test_view_with_all_callbacks(self, ws_server):
        view = ws_server.create_route("ws/full/", "full_group")
        events = []

        @view.connect("before")
        def on_connect(conn):
            events.append(("connect", conn.path))

        @view.receive
        def on_receive(conn, data):
            events.append(("receive", data))

        @view.disconnect
        def on_disconnect(conn, code=None, reason=None):
            events.append(("disconnect", code))

        # Verify all callbacks are registered and callable
        conn = Connection("/ws/full/", "", {}, {})
        on_connect(conn)
        on_receive(conn, "test message")
        on_disconnect(conn, 1000)

        assert events == [
            ("connect", "/ws/full/"),
            ("receive", "test message"),
            ("disconnect", 1000),
        ]

    def test_view_callbacks_access_conn_data(self, ws_server):
        view = ws_server.create_route("ws/conn_access/", "conn_access")
        captured_data = {}

        @view.connect("before")
        def on_connect(conn):
            captured_data["path"] = conn.path
            captured_data["query"] = conn.query_string
            captured_data["cookie"] = conn.get_cookie("session")
            captured_data["header"] = conn.get_header("authorization")

        conn = Connection(
            path="/ws/conn_access/",
            query_string="room=general",
            headers={"authorization": "Bearer token123"},
            cookies={"session": "abc123"},
        )
        on_connect(conn)

        assert captured_data["path"] == "/ws/conn_access/"
        assert captured_data["query"] == "room=general"
        assert captured_data["cookie"] == "abc123"
        assert captured_data["header"] == "Bearer token123"



import json

from webrockets.client import connect
from webrockets.django import server
from webrockets.django.test import WebsocketTestCase


class ChatTestCase(WebsocketTestCase):
    def test_connect_receives_welcome_message(self):
        """Test that connecting sends a welcome system message."""
        with connect(f"ws://{server.addr()}/ws/chat/") as ws:
            response = json.loads(ws.recv())
            self.assertEqual(response["type"], "system")
            self.assertIn("Connected", response["message"])

    def test_join_room(self):
        """Test joining a chat room."""
        with connect(f"ws://{server.addr()}/ws/chat/") as ws:
            ws.recv()  # welcome message
            ws.send(json.dumps({"type": "join", "room": "test", "username": "alice"}))
            response = json.loads(ws.recv())
            self.assertEqual(response["type"], "joined")
            self.assertEqual(response["room"], "test")
            self.assertEqual(response["message"], "You joined test")

    def test_leave_room(self):
        """Test leaving a chat room."""
        with connect(f"ws://{server.addr()}/ws/chat/") as ws:
            ws.recv()  # welcome message
            ws.send(json.dumps({"type": "join", "room": "test"}))
            ws.recv()  # joined confirmation
            ws.recv()  # user_joined broadcast

            ws.send(json.dumps({"type": "leave"}))
            response = json.loads(ws.recv())
            self.assertEqual(response["type"], "left")
            self.assertEqual(response["room"], "test")

    def test_leave_without_room_returns_error(self):
        """Test that leaving without being in a room returns an error."""
        with connect(f"ws://{server.addr()}/ws/chat/") as ws:
            ws.recv()  # welcome message
            ws.send(json.dumps({"type": "leave"}))
            response = json.loads(ws.recv())
            self.assertEqual(response["type"], "error")
            self.assertIn("not in any room", response["message"])

    def test_send_message_to_room(self):
        """Test sending a message broadcasts to room members."""
        with connect(f"ws://{server.addr()}/ws/chat/") as ws1:
            with connect(f"ws://{server.addr()}/ws/chat/") as ws2:
                # Both clients connect and join the same room
                ws1.recv()  # welcome
                ws2.recv()  # welcome

                ws1.send(json.dumps({"type": "join", "room": "chat", "username": "alice"}))
                ws1.recv()  # joined
                ws1.recv()  # user_joined broadcast

                ws2.send(json.dumps({"type": "join", "room": "chat", "username": "bob"}))
                ws2.recv()  # joined
                ws1.recv()  # user_joined for bob
                ws2.recv()  # user_joined broadcast

                # Alice sends a message
                ws1.send(json.dumps({"type": "message", "message": "Hello everyone!"}))

                # Both should recv the broadcast
                msg1 = json.loads(ws1.recv())
                msg2 = json.loads(ws2.recv())

                self.assertEqual(msg1["type"], "message")
                self.assertEqual(msg1["sender"], "alice")
                self.assertEqual(msg1["message"], "Hello everyone!")

                self.assertEqual(msg2["type"], "message")
                self.assertEqual(msg2["sender"], "alice")

    def test_message_without_room_returns_error(self):
        """Test that sending a message without joining a room returns an error."""
        with connect(f"ws://{server.addr()}/ws/chat/") as ws:
            ws.recv()  # welcome message
            ws.send(json.dumps({"type": "message", "message": "hello"}))
            response = json.loads(ws.recv())
            self.assertEqual(response["type"], "error")
            self.assertIn("Join a room first", response["message"])

    def test_unknown_message_type_returns_error(self):
        """Test that unknown message types return an error."""
        with connect(f"ws://{server.addr()}/ws/chat/") as ws:
            ws.recv()  # welcome message
            ws.send(json.dumps({"type": "unknown_type"}))
            response = json.loads(ws.recv())
            self.assertEqual(response["type"], "error")
            self.assertIn("Unknown message type", response["message"])

    def test_invalid_json_returns_error(self):
        """Test that invalid JSON returns an error."""
        with connect(f"ws://{server.addr()}/ws/chat/") as ws:
            ws.recv()  # welcome message
            ws.send("not valid json")
            response = json.loads(ws.recv())
            self.assertEqual(response["type"], "error")



from types import SimpleNamespace
from typing import Any
from urllib.parse import parse_qs

from django.conf import settings
from django.contrib import auth
from django.utils.module_loading import import_string

from webrockets import IncomingConnection
from webrockets.auth import AuthenticationFailed, BaseAuthentication


class SessionAuthentication(BaseAuthentication):
    def __init__(self, session_cookie_name: str | None = None):
        """
        Initialize SessionAuthentication.

        Args:
            session_cookie_name: Name of the session cookie. If None, uses
                                 Django's SESSION_COOKIE_NAME setting.
        """
        self._session_cookie_name = session_cookie_name

    @property
    def session_cookie_name(self) -> str:
        if self._session_cookie_name:
            return self._session_cookie_name
        return getattr(settings, "SESSION_COOKIE_NAME", "sessionid")

    def _get_session_store(self):
        engine = getattr(settings, "SESSION_ENGINE", "django.contrib.sessions.backends.db")
        return import_string(f"{engine}.SessionStore")

    def authenticate(self, conn: IncomingConnection) -> Any | None:
        session_id = conn.get_cookie(self.session_cookie_name)
        if not session_id:
            raise AuthenticationFailed("No session cookie found", close_code=4001)

        SessionStore = self._get_session_store()
        session = SessionStore(session_key=session_id)

        request = SimpleNamespace(session=session)
        user = auth.get_user(request)
        if not user.is_authenticated:
            raise AuthenticationFailed("No authenticated user in session", close_code=4001)

        return user


class CookieTokenAuthentication(BaseAuthentication):
    """
    Authentication class that reads a token from a cookie.

    This is useful for JWT tokens stored in HTTP-only cookies.
    Users must implement the `validate_token` method or provide a validator function.

    Example:
        class JWTCookieAuth(CookieTokenAuthentication):
            cookie_name = "jwt_token"

            def validate_token(self, token: str) -> Any:
                payload = jwt.decode(token, settings.SECRET_KEY, algorithms=["HS256"])
                return User.objects.get(pk=payload["user_id"])
    """

    cookie_name: str = "auth_token"

    def authenticate(self, conn: IncomingConnection) -> Any | None:
        token = conn.get_cookie(self.cookie_name)
        if not token:
            raise AuthenticationFailed(f"No {self.cookie_name} cookie found", close_code=4001)

        user = self.validate_token(token)
        return user

    def validate_token(self, token: str) -> Any:
        """
        Validate the token and return the user.

        Override this method in subclasses to implement token validation.

        Args:
            token: The token string from the cookie.

        Returns:
            The authenticated user object.

        Raises:
            AuthenticationFailed: If the token is invalid.
        """
        raise NotImplementedError("Subclasses must implement validate_token()")


class HeaderTokenAuthentication(BaseAuthentication):
    """
    Authentication class that reads a token from an HTTP header.

    Note: WebSocket connections from browsers cannot set custom headers
    in the initial handshake. This is primarily useful for non-browser clients
    or when using query parameters as a fallback.

    For browser-based JWT authentication, consider:
    - Using CookieTokenAuthentication with HTTP-only cookies
    - Passing tokens via query string (QueryStringTokenAuthentication)
    - Performing authentication after connection via message protocol

    Example for non-browser clients:
        class BearerTokenAuth(HeaderTokenAuthentication):
            header_name = "authorization"
            keyword = "Bearer"

            def validate_token(self, token: str) -> Any:
                return verify_and_get_user(token)
    """

    header_name: str = "authorization"
    keyword: str = "Bearer"

    def authenticate(self, conn: IncomingConnection) -> Any | None:
        auth_header = conn.get_header(self.header_name)
        if not auth_header:
            raise AuthenticationFailed(f"No {self.header_name} header found", close_code=4001)

        parts = auth_header.split()
        if len(parts) != 2 or parts[0].lower() != self.keyword.lower():
            raise AuthenticationFailed(f"Invalid {self.header_name} header format", close_code=4001)

        token = parts[1]
        user = self.validate_token(token)
        return user

    def validate_token(self, token: str) -> Any:
        """
        Validate the token and return the user.

        Override this method in subclasses to implement token validation.
        """
        raise NotImplementedError("Subclasses must implement validate_token()")


class QueryStringTokenAuthentication(BaseAuthentication):
    """
    Authentication class that reads a token from the query string.

    This is useful for browser-based WebSocket connections where custom headers
    cannot be set. The token is passed as a URL parameter.

    Example URL: ws://example.com/chat/?token=your-jwt-token

    Note: Query string tokens may be logged in server access logs.
    Consider using short-lived tokens for this authentication method.

    Example:
        class JWTQueryAuth(QueryStringTokenAuthentication):
            query_param = "token"

            def validate_token(self, token: str) -> Any:
                payload = jwt.decode(token, settings.SECRET_KEY, algorithms=["HS256"])
                return User.objects.get(pk=payload["user_id"])
    """

    query_param: str = "token"

    def authenticate(self, conn: IncomingConnection) -> Any | None:
        params = parse_qs(conn.query_string)
        tokens = params.get(self.query_param, [])

        if not tokens:
            raise AuthenticationFailed(f"No {self.query_param} in query string", close_code=4001)

        token = tokens[0]
        user = self.validate_token(token)
        return user

    def validate_token(self, token: str) -> Any:
        """
        Validate the token and return the user.

        Override this method in subclasses to implement token validation.
        """
        raise NotImplementedError("Subclasses must implement validate_token()")



# ruff: noqa: E402, I001
# Import order matters: webrockets must be imported first to avoid circular imports
from typing import Literal, TypedDict
from .webrockets import *
from .utils import noop
from .auth import BaseAuthentication


class _RedisBrokerConfigOptional(TypedDict, total=False):
    url: str  # default: "redis://localhost:6379"
    channel: str  # default: "ws_broadcast"


class RedisBrokerConfig(_RedisBrokerConfigOptional):
    type: Literal["redis"]


class _AmqpBrokerConfigOptional(TypedDict, total=False):
    url: str  # default: "amqp://localhost:5672"
    exchange: str  # default: "ws_broadcast"
    queue: str | None  # default: auto-generated UUID
    routing_key: str  # default: "#"


class AmqpBrokerConfig(_AmqpBrokerConfigOptional):
    type: Literal["amqp"]


BrokerConfig = RedisBrokerConfig | AmqpBrokerConfig

__all__ = ["noop", "BaseAuthentication", "BrokerConfig"]

__doc__ = webrockets.__doc__
if hasattr(webrockets, "__all__"):
    __all__ += webrockets.__all__



import asyncio

from websockets.asyncio.server import serve


async def router(websocket):
    async for message in websocket:
        await websocket.send(message)


async def main():
    async with serve(router, "0.0.0.0", 4200, backlog=100_000) as server:
        await server.serve_forever()


if __name__ == "__main__":
    asyncio.run(main())



from abc import ABC, abstractmethod
from typing import Any

from webrockets import IncomingConnection


class AuthenticationFailed(Exception):
    """
    Exception raised when authentication fails.
    """

    def __init__(self, detail: str = "Authentication failed", close_code: int = 4003):
        self.detail = detail
        self.close_code = close_code
        super().__init__(detail)


class BaseAuthentication(ABC):
    """
    Base class for WebSocket authentication.

    All authentication classes should extend this class and implement
    the `authenticate` method.

    Example usage:
        class TokenAuthentication(BaseAuthentication):
            def authenticate(self, conn: IncommingConnection) -> Any:
                token = conn.get_header("Authorization")
                if not token:
                    raise AuthenticationFailed("No token provided")
                user = validate_token(token)
                return user
    """

    @abstractmethod
    def authenticate(self, conn: IncomingConnection) -> Any | None:
        """
        Authenticate the WebSocket connection request.

        Args:
            conn: IncommingConnection containing headers, cookies, path, and query string.

        Returns:
            A user if authentication succeeds.
            Return None to skip this authenticator and try the next one.

        Raises:
            AuthenticationFailed: If authentication explicitly fails.
        """
        pass



import pytest
from webrockets.auth import AuthenticationFailed
from webrockets.django.auth import (
    CookieTokenAuthentication,
    HeaderTokenAuthentication,
    QueryStringTokenAuthentication,
    SessionAuthentication,
)


@pytest.mark.django_db
class TestSessionAuthentication:
    """Tests for SessionAuthentication."""

    def test_no_session_cookie(self, websocket_scope):
        auth = SessionAuthentication()
        scope = websocket_scope(cookies={})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "No session cookie found" in exc_info.value.detail
        assert exc_info.value.close_code == 4001

    def test_invalid_session_id(self, websocket_scope):
        auth = SessionAuthentication()
        scope = websocket_scope(cookies={"sessionid": "invalid-session-id-12345"})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert exc_info.value.close_code == 4001

    def test_valid_session(self, websocket_scope, create_session, active_user):
        session = create_session(user=active_user)
        auth = SessionAuthentication()
        scope = websocket_scope(cookies={"sessionid": session.session_key})

        result = auth.authenticate(scope)

        assert result is not None
        assert result.pk == active_user.pk

    def test_inactive_user_rejected_by_default(self, websocket_scope, session_store, inactive_user):
        session = session_store()
        session["_auth_user_id"] = str(inactive_user.pk)
        session["_auth_user_backend"] = "django.contrib.auth.backends.ModelBackend"
        session.create()

        auth = SessionAuthentication()
        scope = websocket_scope(cookies={"sessionid": session.session_key})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "No authenticated user in session" in exc_info.value.detail

    def test_custom_session_cookie_name(self, websocket_scope, create_session, active_user):
        session = create_session(user=active_user)
        auth = SessionAuthentication(session_cookie_name="my_session")

        scope = websocket_scope(cookies={"sessionid": session.session_key})
        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)
        assert "No session cookie found" in exc_info.value.detail

        scope = websocket_scope(cookies={"my_session": session.session_key})
        result = auth.authenticate(scope)
        assert result is not None
        assert result.pk == active_user.pk

    def test_session_without_user_id(self, websocket_scope, session_store):
        session = session_store()
        session["some_data"] = "value"
        session.create()

        auth = SessionAuthentication()
        scope = websocket_scope(cookies={"sessionid": session.session_key})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "No authenticated user in session" in exc_info.value.detail

    def test_user_not_found(self, websocket_scope, session_store):
        session = session_store()
        session["_auth_user_id"] = "99999"
        session["_auth_user_backend"] = "django.contrib.auth.backends.ModelBackend"
        session.create()

        auth = SessionAuthentication()
        scope = websocket_scope(cookies={"sessionid": session.session_key})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "No authenticated user in session" in exc_info.value.detail

    def test_session_auth_hash_validation(self, websocket_scope, session_store, active_user):
        # Create session with wrong auth hash (simulates password change)
        session = session_store()
        session["_auth_user_id"] = str(active_user.pk)
        session["_auth_user_backend"] = "django.contrib.auth.backends.ModelBackend"
        session["_auth_user_hash"] = "wrong-hash-after-password-change"
        session.create()

        auth = SessionAuthentication()
        scope = websocket_scope(cookies={"sessionid": session.session_key})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "No authenticated user in session" in exc_info.value.detail


class TestCookieTokenAuthentication:
    """Tests for CookieTokenAuthentication."""

    def test_no_cookie(self, websocket_scope):
        class TestAuth(CookieTokenAuthentication):
            cookie_name = "auth_token"

            def validate_token(self, token):
                return "user"

        auth = TestAuth()
        scope = websocket_scope(cookies={})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "No auth_token cookie found" in exc_info.value.detail

    def test_valid_token(self, websocket_scope):
        class TestAuth(CookieTokenAuthentication):
            cookie_name = "jwt"

            def validate_token(self, token):
                if token == "valid-token":
                    return {"id": 1, "username": "testuser"}
                raise AuthenticationFailed("Invalid token")

        auth = TestAuth()
        scope = websocket_scope(cookies={"jwt": "valid-token"})

        result = auth.authenticate(scope)

        assert result is not None
        assert result == {"id": 1, "username": "testuser"}

    def test_invalid_token(self, websocket_scope):
        class TestAuth(CookieTokenAuthentication):
            cookie_name = "jwt"

            def validate_token(self, token):
                raise AuthenticationFailed("Token expired", close_code=4001)

        auth = TestAuth()
        scope = websocket_scope(cookies={"jwt": "expired-token"})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "Token expired" in exc_info.value.detail

    def test_validate_token_not_implemented(self, websocket_scope):
        auth = CookieTokenAuthentication()
        scope = websocket_scope(cookies={"auth_token": "some-token"})

        with pytest.raises(NotImplementedError):
            auth.authenticate(scope)


class TestHeaderTokenAuthentication:
    """Tests for HeaderTokenAuthentication."""

    def test_no_header(self, websocket_scope):
        class TestAuth(HeaderTokenAuthentication):
            def validate_token(self, token):
                return "user"

        auth = TestAuth()
        scope = websocket_scope(headers={})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "No authorization header found" in exc_info.value.detail

    def test_invalid_header_format_no_keyword(self, websocket_scope):
        class TestAuth(HeaderTokenAuthentication):
            def validate_token(self, token):
                return "user"

        auth = TestAuth()
        scope = websocket_scope(headers={"authorization": "token123"})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "Invalid authorization header format" in exc_info.value.detail

    def test_invalid_header_format_wrong_keyword(self, websocket_scope):
        class TestAuth(HeaderTokenAuthentication):
            keyword = "Bearer"

            def validate_token(self, token):
                return "user"

        auth = TestAuth()
        scope = websocket_scope(headers={"authorization": "Basic token123"})

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "Invalid authorization header format" in exc_info.value.detail

    def test_valid_bearer_token(self, websocket_scope):
        class TestAuth(HeaderTokenAuthentication):
            def validate_token(self, token):
                return {"user_id": 1}

        auth = TestAuth()
        scope = websocket_scope(headers={"authorization": "Bearer my-jwt-token"})

        result = auth.authenticate(scope)

        assert result is not None
        assert result == {"user_id": 1}

    def test_custom_header_and_keyword(self, websocket_scope):
        class TestAuth(HeaderTokenAuthentication):
            header_name = "x-api-key"
            keyword = "Key"

            def validate_token(self, token):
                return {"api_key": token}

        auth = TestAuth()
        scope = websocket_scope(headers={"x-api-key": "Key abc123"})

        result = auth.authenticate(scope)

        assert result is not None
        assert result == {"api_key": "abc123"}


class TestQueryStringTokenAuthentication:
    """Tests for QueryStringTokenAuthentication."""

    def test_no_token_in_query(self, websocket_scope):
        class TestAuth(QueryStringTokenAuthentication):
            def validate_token(self, token):
                return "user"

        auth = TestAuth()
        scope = websocket_scope(query_string="room=general")

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "No token in query string" in exc_info.value.detail

    def test_empty_query_string(self, websocket_scope):
        class TestAuth(QueryStringTokenAuthentication):
            def validate_token(self, token):
                return "user"

        auth = TestAuth()
        scope = websocket_scope(query_string="")

        with pytest.raises(AuthenticationFailed) as exc_info:
            auth.authenticate(scope)

        assert "No token in query string" in exc_info.value.detail

    def test_valid_token_in_query(self, websocket_scope):
        class TestAuth(QueryStringTokenAuthentication):
            def validate_token(self, token):
                if token == "valid-jwt":
                    return {"user_id": 42}
                raise AuthenticationFailed("Invalid")

        auth = TestAuth()
        scope = websocket_scope(query_string="token=valid-jwt&room=chat")

        result = auth.authenticate(scope)

        assert result is not None
        assert result == {"user_id": 42}

    def test_custom_query_param(self, websocket_scope):
        class TestAuth(QueryStringTokenAuthentication):
            query_param = "access_token"

            def validate_token(self, token):
                return {"token": token}

        auth = TestAuth()
        scope = websocket_scope(query_string="access_token=abc123")

        result = auth.authenticate(scope)

        assert result is not None
        assert result == {"token": "abc123"}

    def test_url_encoded_token(self, websocket_scope):
        class TestAuth(QueryStringTokenAuthentication):
            def validate_token(self, token):
                return token

        auth = TestAuth()
        scope = websocket_scope(query_string="token=hello%20world")

        result = auth.authenticate(scope)

        assert result is not None
        assert result == "hello world"



from webrockets import Connection, WebsocketServer

server = WebsocketServer(host="0.0.0.0", port=6969)

echo = server.create_route(path="echo", default_group="echo")


@echo.receive
def echo_receive(conn: Connection, data: str | bytes):
    conn.send(data)


if __name__ == "__main__":
    server.start()
